@@ -1,5 +1,6 @@
 import base64
 import copy
+import glob
 import importlib
 import importlib.util as util
 import inspect
@@ -8,6 +9,7 @@ import os
 import site
 import socket
 import sqlite3
+import sys
 import winreg
 import xml.etree.ElementTree as ElTree
 import zipfile
@@ -17,9 +19,25 @@ from inspect import signature
 from shutil import copyfile
 from typing import List, Any
 from urllib import parse
+
 import xmltodict
 
 
+# The BPMN-RPA WorkflowEngine is free software: you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# The BPMN-RPA WorkflowEngine is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program.  If not, see <https://www.gnu.org/licenses/>.
+# Copyright 2020-2021 Joost van Gils (J.W.N.M. van Gils)
+
+
 class WorkflowEngine:
 
     def __init__(self, input_parameter: Any = None, pythonpath: str = "", installation_directory: str = ""):
@@ -104,14 +122,15 @@ class WorkflowEngine:
         self.step_nr = None
         self.variables = {}  # Dictionary to hold WorkflowEngine variables
 
-    def get_input_parameter(self, as_dictionary: False):
+    def get_input_parameter(self, as_dictionary: bool = False) -> Any:
         """
         Returns the input parameter that was given when creating an instance of the WorkflowEngine
         :param as_dictionary: Optional. Indicator whether to treat the given input as a dictionary object (string to dict).
         :return: The input_parameter that was given when creating an instance of the WorkflowEngine
         """
         if as_dictionary:
-            self.input_parameter = eval(self.input_parameter)
+            if not isinstance(self.input_parameter, dict):
+                self.input_parameter = eval(self.input_parameter)
         self.print_log(f"Got input parameter {str(self.input_parameter)}")
         return self.input_parameter
 
@@ -127,16 +146,25 @@ class WorkflowEngine:
             self.flowpath = filepath
         xml_file = open(filepath, "r")
         if not filepath.__contains__(".vsdx"):
-            self.flowname = filepath.split("\\")[-1].lower().replace(".xml", "")
-            xml_root = ElTree.fromstring(xml_file.read())
-            raw_text = xml_root[0].text
-            base64_decode = base64.b64decode(raw_text)
-            inflated_xml = zlib.decompress(base64_decode, -zlib.MAX_WBITS).decode("utf-8")
-            url_decode = parse.unquote(inflated_xml)
-            if as_xml:
-                return url_decode
-            retn = xmltodict.parse(url_decode)
-            return retn
+            if filepath.__contains__(".flw"):
+                self.flowname = filepath.split("\\")[-1].lower().replace(".flw", "")
+                with open(filepath, encoding="utf8", errors='ignore') as f:
+                    content = f.read()
+                decoded = base64.b64decode(content).decode("utf-8", errors='ignore')
+                idx = decoded.index("}]")
+                dict_list = json.loads(decoded[0:idx + 2])
+                return dict_list
+            else:
+                self.flowname = filepath.split("\\")[-1].lower().replace(".xml", "")
+                xml_root = ElTree.fromstring(xml_file.read())
+                raw_text = xml_root[0].text
+                base64_decode = base64.b64decode(raw_text)
+                inflated_xml = zlib.decompress(base64_decode, -zlib.MAX_WBITS).decode("utf-8")
+                url_decode = parse.unquote(inflated_xml)
+                if as_xml:
+                    return url_decode
+                retn = xmltodict.parse(url_decode)
+                return retn
         else:
             # It is a MsVisio file!
             visio = Visio()
@@ -218,6 +246,14 @@ class WorkflowEngine:
             # It is a Visio Object!
             visio = ordered_dict
             return visio.get_flow()
+        if isinstance(ordered_dict, list):
+            retn = []
+            for rw in ordered_dict:
+                tmp = self.dynamic_object()
+                for k, v in rw.items():
+                    setattr(tmp, k, v)
+                retn.append(tmp)
+            return retn
         connectors = []
         shapes = []
         connectorvalues = {}
@@ -611,8 +647,9 @@ class WorkflowEngine:
                     self.step_name = step.name
                     if len(step.name) == 0:
                         if hasattr(step, "type"):
-                            self.print_log(status="Running",
-                                           result=f"Passing an {step.type} with value {output_previous_step}...")
+                            if not hasattr(step, "function"):
+                                self.print_log(status="Running",
+                                               result=f"Passing an {step.type} with value {output_previous_step}...")
                     else:
                         if hasattr(step, "function"):
                             if step.function != "print_log":
@@ -1090,6 +1127,14 @@ class WorkflowEngine:
     class dynamic_object(object):
         pass
 
+    @staticmethod
+    def set_breakpoint():
+        """
+        Set a breakpoint to debug the code.
+        """
+        print("---------- Debug ----------")
+        breakpoint()
+
 
 class SQL:
 
@@ -1263,9 +1308,10 @@ class Visio:
                                         return shape
                                 else:
                                     for rw in sec["Row"]:
-                                        if {"@N": "Label", "@V": "Type"} in rw.get("Cell") and {"@N": "Value", "@V": "Start", "@U": "STR"} in rw.get("Cell"):
-                                            setattr(shape, "IsStart", True)
-                                            return shape
+                                        if "Cell" in rw:
+                                            if {"@N": "Label", "@V": "Type"} in rw.get("Cell") and {"@N": "Value", "@V": "Start", "@U": "STR"} in rw.get("Cell"):
+                                                setattr(shape, "IsStart", True)
+                                                return shape
                 count += 1
             else:
                 break
@@ -1280,47 +1326,26 @@ class Visio:
         count = 1
         while True:
             if f"visio/pages/page{count}.xml" in self.root:
-                for shp in self.root[f"visio/pages/page{count}.xml"].get("PageContents").get("Connects").get("Connect"):
-                    blnupdate = False
-                    if [x for x in retn if x["@ID"] == shp["@FromSheet"]]:
-                        conn = [x for x in retn if x["@ID"] == shp["@FromSheet"]][0]
-                        blnupdate = True
-                    else:
+                connects = self.root[f"visio/pages/page{count}.xml"].get("PageContents").get("Connects").get("Connect")
+                if connects:
+                    for shp in self.root[f"visio/pages/page{count}.xml"].get("PageContents").get("Connects").get("Connect"):
                         conn = {"@ID": shp["@FromSheet"], "type": "connector"}
-                        obj = [x for x in self.root[f"visio/pages/page{count}.xml"].get("PageContents").get("Shapes").get("Shape") if x["@ID"] == shp["@FromSheet"]][0]
-                        text = None
-                        if "Text" in obj:
-                            text = obj["Text"]
-                        if text is not None:
-                            conn.update({"value": text})
-                    if shp["@FromCell"] == "EndX":
-                        conn.update({"target": shp["@ToSheet"]})
-                    else:
-                        conn.update({"source": shp["@ToSheet"]})
-                    if not blnupdate:
-                        retn.append(conn)
-                count += 1
+                        if shp["@FromCell"] == "BeginX":
+                            obj = [x for x in self.root[f"visio/pages/page{count}.xml"].get("PageContents").get("Shapes").get("Shape") if x["@ID"] == shp["@FromSheet"]][0]
+                            text = None
+                            if "Text" in obj:
+                                text = obj["Text"]
+                            if text is not None:
+                                conn.update({"value": text})
+                            conn.update({"source": shp["@ToSheet"]})
+                            target = [y for y in [x for x in connects if x["@FromSheet"] == shp["@FromSheet"]] if y["@FromCell"] == "EndX"][0]["@ToSheet"]
+                            conn.update({"target": target})
+                            retn.append(conn)
+                    count += 1
             else:
                 break
         return retn
 
-    def check_dimensions(self, shape: Any) -> Any:
-        """
-        Get Width and Height attributes from master if not already in shape attributes.
-        :param shape: The shape object to check.
-        :return: The shape object including height and width attributes.
-        """
-        if [x for x in shape["Cell"] if x["@N"] == "Width"] and [x for x in shape["Cell"] if x["@N"] == "Height"]:
-            return shape
-        master = [x for x in self.root[f"visio/masters/masters.xml"].get("Masters").get("Master") if x["@ID"] == shape["@Master"]][0]
-        rel = master["Rel"].get("@r:id")
-        target = [x for x in self.root[f"visio/masters/_rels/masters.xml.rels"].get("Relationships").get("Relationship") if x["@Id"] == rel][0]["@Target"]
-        w = [x for x in self.root[f"visio/masters/{target}"].get("MasterContents").get("Shapes").get("Shape").get("Cell") if x["@N"] == "Width"][0]
-        h = [x for x in self.root[f"visio/masters/{target}"].get("MasterContents").get("Shapes").get("Shape").get("Cell") if x["@N"] == "Height"][0]
-        shape["Cell"].append(w)
-        shape["Cell"].append(h)
-        return shape
-
     def get_flow(self):
         """
         Get the flow from the file.
@@ -1328,7 +1353,6 @@ class Visio:
         """
         flow_ = []
         shape = self.get_start()
-        shape = self.check_dimensions(shape)
         flow_.append(shape)
         count = 1
         while True:
@@ -1348,8 +1372,9 @@ class Visio:
         flow_ += connectors
         retn = []
         for s in flow_:
-            step = self.get_step_from_shape(s)
-            retn.append(step)
+            if s is not None:
+                step = self.get_step_from_shape(s)
+                retn.append(step)
         return retn
 
     def get_step_from_shape(self, shape: Any) -> Any:
@@ -1358,30 +1383,46 @@ class Visio:
         :param shape: The Shape-object
         :returns: A Step-object
         """
+        properties = {"id": shape.get("@ID")}
+        properties.update({"type": "shape"})
+        if "@Name" in shape:
+            properties.update({"name": shape["@Name"]})
+        properties.update({"IsStart": False})
         retn = self.dynamic_object()
-        retn.id = shape.get("@ID")
-        retn.type = "Shape"
-        retn.IsStart = False
+        if "type" in shape:
+            if str(shape["type"]).lower().__contains__("connector"):
+                properties.pop("IsStart")
+                properties.update({"type": "connector"})
+                properties.update({"target": shape["target"]})
+                properties.update({"source": shape["source"]})
+                if "value" in shape:
+                    properties.update({"value": shape["value"]})
+                    properties.update({"name": shape["value"]})
+                for k, v in properties.items():
+                    setattr(retn, k, v)
+                return retn
+
         # Get default attributes if any
-        master = None
         standard_attributes = []
+        props = []
         target_id = None
+        # region attributes from template
         if "@Master" in shape:
             master = shape["@Master"]
-            obj = [x for x in self.root[f"visio/masters/masters.xml"].get("Masters").get("Master") if x["@ID"]==master]
+            obj = [x for x in self.root[f"visio/masters/masters.xml"].get("Masters").get("Master") if x["@ID"] == master]
             if obj:
                 r_id = obj[0]["Rel"]["@r:id"]
                 obj = [x for x in self.root[f"visio/masters/_rels/masters.xml.rels"].get("Relationships").get("Relationship") if x["@Id"] == r_id]
                 if obj:
                     target_id = obj[0]["@Target"]
-                    obj = [x for x in self.root[f"visio/masters/{target_id}"]["MasterContents"]["Shapes"]["Shape"]["Section"] if x["@N"]=="Property"]
+                    obj = [x for x in self.root[f"visio/masters/{target_id}"]["MasterContents"]["Shapes"]["Shape"]["Section"] if x["@N"] == "Property"]
                     if obj:
                         if "Cell" in obj[0]["Row"]:
-                            standard_attributes= obj[0]["Row"]["Cell"]
+                            standard_attributes = obj[0]["Row"]["Cell"]
                         else:
                             for rw in obj[0]["Row"]:
-                                    for attr in rw["Cell"]:
-                                        standard_attributes.append(attr)
+                                for attr in rw["Cell"]:
+                                    standard_attributes.append(attr)
         key = None
         value = None
         for rw in standard_attributes:
@@ -1390,94 +1431,93 @@ class Visio:
             if rw["@N"] == "Value":
                 value = rw["@V"]
             if key is not None and value is not None:
-                setattr(retn, str(key).replace("@", "").lower(), value)
+                properties.update({"row_id": "template", str(key).lower(): value})
                 # reset key and value
                 key, value = None, None
+        labels = []
         if target_id is not None:
-            labels = [x for x in self.root[f"visio/masters/{target_id}"]["MasterContents"]["Shapes"]["Shape"]["Section"] if x["@N"] == "Property"][0]["Row"]
-        else:
-            labels = []
+            obj = [x for x in self.root[f"visio/masters/{target_id}"]["MasterContents"]["Shapes"]["Shape"]["Section"] if x["@N"] == "Property"]
+            if obj:
+                labels = obj[0]["Row"]
+        # endregion
         if "Section" in shape:
             if not isinstance(shape["Section"], list):
-                col = [shape["Section"]]
+                sect = [shape["Section"]]
             else:
-                col = [x for x in shape["Section"] if x["@N"] == "Property"]
-            if [x for x in col if x["@N"] == "Property"]:
-                col = [x for x in col if x["@N"] == "Property"][0].get("Row")
+                sect = [x for x in shape["Section"] if x["@N"] == "Property"]
+            # sect is a list or ordered dict. Get the 'Property' section
+            if [x for x in sect if x["@N"] == "Property"]:
+                props = [x for x in sect if x["@N"] == "Property"][0].get("Row")
+            # props is now the 'Property' section of the Shape
+            # Loop the properties
+            # props is ordered_dict based on stencil, but list[ordered_dict] for manual
+            if not isinstance(props, list):
+                props = [props]
+            for prop in props:
+                # get the property row_id
                 row_id = None
-                if labels:
-                    if isinstance(labels, list):
+                if str(prop.get("@N")).lower().__contains__("row_"):
+                    row_id = prop.get("@N")
+                # region get the key and value
+                key, value = None, None
+                if "Cell" not in prop:
+                    continue
+                if isinstance(prop.get("Cell"), dict):
+                    # it is a value that belongs to a template label
+                    label = [x for x in labels if x["@N"] == row_id]
+                    kvps = None
+                    if label:
+                        kvps = label[0]["Cell"]
+                    rw_k = [x for x in kvps if x["@N"] == "Label"]
+                    if rw_k:
+                        key = rw_k[0]["@V"]
+                    value = prop.get("Cell")["@V"]
+                else:
+                    rw_k = [x for x in prop.get("Cell") if x["@N"] == "Label"]
+                    if rw_k:
+                        key = rw_k[0]["@V"]
+                    rw_l = [x for x in prop.get("Cell") if x["@N"] == "Value"]
+                    if rw_l:
+                        value = rw_l[0]["@V"]
+                    # endregion
+                if key is not None and value is not None:
+                    properties.update({"row_id": row_id, str(key).lower(): value})
 
-                            if "Cell" in col:
-                                row_id = col["@N"]
-                                col = col["Cell"]
-                            if not isinstance(col, list):
-                                test = [x for x in labels if x["@N"] == row_id]
-                                if test:
-                                    label = [y for y in [x for x in labels if x["@N"] == row_id][0]["Cell"] if y["@N"] == "Label"][0]["@V"]
-                                key = label
-                                value = col["@V"]
-                                setattr(retn, str(key).replace("@", "").lower(), value)
-                                # reset key and value
-                                key, value = None, None
-                            else:
-                                for rw in col:
-                                    row_id = rw["@N"]
-                                    if "Cell" in rw:
-                                        rw = rw["Cell"]
-                                    test = [x for x in labels if x["@N"] == row_id]
-                                    if test:
-                                        label = [y for y in [x for x in labels if x["@N"] == row_id][0]["Cell"] if y["@N"] == "Label"][0]["@V"]
-                                    key = label
-                                    if rw["@N"] == "Value":
-                                        value = rw["@V"]
-                                    if key is not None and value is not None:
-                                        if str(key).lower() == "class":
-                                            key = "classname"  # 'class' is a reserved keyword, so use 'classname'
-                                        if str(key).lower() != "name":
-                                            setattr(retn, str(key).replace("@", "").lower(), value)
-                                            # reset key and value
-                                            key, value = None, None
-                                        continue
-            else:
-                for att in shape:
-                    setattr(retn, str(att).replace("@", "").lower(), shape[att])
-        else:
-            for att in shape:
-                setattr(retn, str(att).replace("@", "").lower(), shape[att])
+        if "Section" in shape:
+            for rw in shape["Section"]:
+                if str(rw).lower().__contains__("('@v', 'exclusive gateway')") and str(rw).lower().__contains__("('@n', 'type')"):
+                    properties.update({"type": "exclusive gateway"})
+                    properties.update({"name": ""})
+                if str(rw).lower().__contains__("('@v', 'parallel gateway')") and str(rw).lower().__contains__("('@n', 'type')"):
+                    properties.update({"type": "parallel gateway"})
+                    properties.update({"name": ""})
 
-        if str(shape.get("@NameU")).__contains__("connector"):
-            retn.type = "connector"
-            if not shape["Text"] is None:
-                retn.value = shape["Text"]
-                retn.name = shape["Text"]
-        else:
-            if "Section" in shape:
-                for rw in shape["Section"]:
-                    if str(rw).lower().__contains__("('@v', 'exclusive gateway')") and str(rw).lower().__contains__("('@n', 'type')"):
-                        retn.type = "exclusive gateway"
-                        retn.name = ""
-                    if str(rw).lower().__contains__("('@v', 'parallel gateway')") and str(rw).lower().__contains__("('@n', 'type')"):
-                        retn.type = "parallel gateway"
-                        retn.name = ""
         if "type" not in shape:
             count = 1
             while True:
                 if f"visio/pages/page{count}.xml" in self.root:
                     if not [x for x in self.root[f"visio/pages/page{count}.xml"].get("PageContents").get("Connects").get("Connect") if x["@FromCell"] == "EndX" and x["@ToSheet"] == shape["@ID"]]:
-                        retn.IsStart = True
-                        retn.name = "Start"
-                        retn.type = "shape"
+                        properties.update({"IsStart": True})
+                        properties.update({"name": "Start"})
+                        properties.update({"type": "shape"})
                     if not [x for x in self.root[f"visio/pages/page{count}.xml"].get("PageContents").get("Connects").get("Connect") if x["@FromCell"] == "BeginX" and x["@ToSheet"] == shape["@ID"]]:
-                        retn.name = "End"
-                        retn.type = "shape"
+                        properties.update({"name": "End"})
+                        properties.update({"type": "shape"})
                     count += 1
                 else:
                     break
+        properties.update({"id": shape.get("@ID")})
+        if "Text" in shape:
+            properties.update({"name": str(shape["Text"]).lower()})
+        if str(properties["name"]).lower().__contains__("gateway"):
+            properties.update({"name": ""})
+        properties.update({"type": str(properties["type"]).lower()})
+        for k, v in properties.items():
+            setattr(retn, k, v)
         return retn
 
 # Test
 # engine = WorkflowEngine()
-# doc = engine.open(fr"c:\\temp\\test2.vsdx")  # c:\\temp\\test.xml
+# doc = engine.open(fr"c:\\temp\\test.vsdx")  # c:\\temp\\test.xml
 # steps = engine.get_flow(doc)
 # engine.run_flow(steps)
